@@ -1,10 +1,20 @@
-from datetime import datetime
+import binascii
+import re
+import sys
 import unicodedata
+
+from codecs import decode, encode
+from datetime import datetime
+
 from anticipate.adapt import adapt, AdaptError
+from six import integer_types, raise_from, string_types, text_type
+from six import reraise as raise_
+from six.moves.urllib.parse import urlparse, urlunparse
+
 from springfield.timeutil import date_parse, generate_rfc3339
 from springfield.types import Empty
 from decimal import Decimal
-import re
+
 
 class FieldDescriptor(object):
     """
@@ -40,6 +50,7 @@ class FieldDescriptor(object):
         if new_value != old_value:
             instance.__changes__.add(self.name)
 
+
 class Field(object):
     """
     A field
@@ -117,6 +128,7 @@ class Field(object):
         """
         return value
 
+
 class AdaptableTypeField(Field):
     """
     A :class:`Field` that has a specific type and can be adapted
@@ -141,7 +153,7 @@ class AdaptableTypeField(Field):
             # Use an object's own adapter to adapt.
             try:
                 return value.__adapt__(self.type)
-            except TypeError as e:
+            except TypeError:
                 pass
 
         if hasattr(self.type, '__adapt__'):
@@ -162,7 +174,6 @@ class AdaptableTypeField(Field):
 
         raise TypeError('Could not adapt %r to %r' % (value, self.type))
 
-
     @classmethod
     def register_adapter(cls, from_cls, func):
         """
@@ -176,11 +187,13 @@ class AdaptableTypeField(Field):
 
         cls.__adapters__[from_cls] = func
 
+
 class IntField(AdaptableTypeField):
     """
     A :class:`Field` that contains an `int`.
     """
     type = int
+
     def adapt(self, value):
         """
         Adapt `value` to an `int`.
@@ -194,20 +207,22 @@ class IntField(AdaptableTypeField):
         try:
             return super(IntField, self).adapt(value)
         except TypeError:
-            if isinstance(value, basestring):
+            if isinstance(value, string_types):
                 return int(value)
-            elif isinstance(value, (float, long)):
+            elif isinstance(value, (float,) + integer_types):
                 t = int(value)
                 if t == value:
                     return t
 
             raise
 
+
 class FloatField(AdaptableTypeField):
     """
     A :class:`Field` that contains a `float`.
     """
     type = float
+
     def adapt(self, value):
         """
         Adapt `value` to a `float`.
@@ -220,21 +235,22 @@ class FloatField(AdaptableTypeField):
         try:
             return super(FloatField, self).adapt(value)
         except TypeError:
-            if isinstance(value, basestring):
+            if isinstance(value, (int,) + string_types):
                 return float(value)
-            elif isinstance(value, (float, long)):
+            elif isinstance(value, (float,) + integer_types):
                 return value
             elif isinstance(value, Decimal):
                 return float(value)
 
             raise
 
+
 class BooleanField(AdaptableTypeField):
     """
     A :class:`Field` that contains a `bool`.
     """
-
     type = bool
+
     def adapt(self, value):
         """
         Adapt `value` to a `bool`.
@@ -254,13 +270,13 @@ class BooleanField(AdaptableTypeField):
         try:
             return super(BooleanField, self).adapt(value)
         except TypeError:
-            if isinstance(value, basestring):
+            if isinstance(value, string_types):
                 str = value.lower()
                 if str in ['yes', 'true', '1', 'on']:
                     return True
                 elif str in ['no', 'false', '0', 'off']:
                     return False
-            elif isinstance(value, (float, long, int)):
+            elif isinstance(value, (float,) + integer_types):
                 if value == 1:
                     return True
                 elif value == 0:
@@ -268,12 +284,13 @@ class BooleanField(AdaptableTypeField):
 
             raise
 
+
 class StringField(AdaptableTypeField):
     """
     A :class:`Field` that contains a unicode string.
     """
+    type = text_type
 
-    type = unicode
     def adapt(self, value):
         """
         Adapt `value` to `unicode`.
@@ -281,10 +298,84 @@ class StringField(AdaptableTypeField):
         try:
             return super(StringField, self).adapt(value)
         except TypeError:
-            if isinstance(value, basestring):
-                return unicode(value)
+            if isinstance(value, string_types):
+                return text_type(value)
             raise
 
+
+class BytesField(AdaptableTypeField):
+    """
+    A :class:`Field` that contains binary `bytes`.
+
+    The field has an encoding to use for json/unicode conversion, such
+    as `'base64'` (the default) or `'hex'`.
+
+    If `encoding == None`, no encoding/decoding is performed for JSON/unicode
+    values which mean JSON itself will have to escape the bytes using unicode
+    escapes where necessary.  This is most suitable for cases where the "bytes"
+    are known to be ASCII 7-bit safe.
+
+    The encoding is used in `adapt` if the input is a `unicode` instance, and
+    in `jsonify` always.
+    """
+
+    type = bytes
+    encoding = 'base64'
+
+    def __init__(self, encoding='base64', *args, **kwargs):
+        """
+        Construct a Bytes field
+
+        :param encoding: Optional encoding to use for jsonify(), such as
+           'hex' or 'base64'
+        """
+        super(BytesField, self).__init__(*args, **kwargs)
+        self.encoding = encoding
+
+    def jsonify(self, value):
+        """
+        Encode the bytes into a unicode string suitable for json encoding.
+
+        If an encoding was specified for the field, it is applied here.
+        """
+        if value is None:
+            return None
+
+        if not isinstance(value, bytes):
+            raise ValueError('BytesField must contain bytes')
+
+        # Apply hex/base64 encoding if desired
+        if self.encoding:
+            value = encode(value, self.encoding)
+
+        # Convert to unicode using an 8-bit encoding to retain binary data
+        return decode(value, 'latin1')
+
+    def adapt(self, value):
+        """
+        If the input is unicode, decode it into bytes.  If it is already
+        bytes, it is returned unchanged.
+
+        If an encoding was specific for the field, it is applied here if the input
+        is `unicode`.
+
+        This assumes that the unicode only contains code points in the
+        valid ranges for a byte - e.g. 0-255.
+
+        :param value: Value to decode
+        :return: `bytes` object
+        """
+        if isinstance(value, text_type):
+            try:
+                value = encode(value, 'latin1')
+                if self.encoding:
+                    value = decode(value, self.encoding)
+            except binascii.Error as e:
+                raise_from(TypeError, e)
+
+        return super(BytesField, self).adapt(value)
+
+
 class SlugField(StringField):
     """
     :class:`Field` whose value is a slugified string.
@@ -305,12 +396,15 @@ class SlugField(StringField):
         if not value:
             return ''
 
-        slug = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
+        # re-encoding to ASCII makes it bytes, and we need unicode,
+        # so we immediately decode
+        slug = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
         slug = re.sub(r'[^\w\s-]', '', slug).strip().lower()
         slug = re.sub(r'[-\s]+', '-', slug).strip('-')
 
         return slug
 
+
 class DateTimeField(AdaptableTypeField):
     """
     :class:`Field` whose value is a Python `datetime.datetime`
@@ -330,7 +424,7 @@ class DateTimeField(AdaptableTypeField):
         try:
             return super(DateTimeField, self).adapt(value)
         except TypeError:
-            if isinstance(value, basestring):
+            if isinstance(value, string_types):
                 return date_parse(value)
             raise
 
@@ -341,32 +435,117 @@ class DateTimeField(AdaptableTypeField):
         if value is not None:
             return generate_rfc3339(value)
 
+
 class EmailField(StringField):
     """
     :class:`Field` with an email value
     """
 
+
 class UrlField(StringField):
     """
     :class:`Field` with a URL value
     """
+    def adapt(self, value):
+        """
+        Validate that the `value` has a valid URL format containing
+        a scheme and network location using urlparse.
+
+        :param value: A url-like value.
+
+        :returns: URL with sheme and network location in lower case.
+        """
+        value = super(UrlField, self).adapt(value)
+        if value:
+            url_parts = urlparse(value)
+            if url_parts.scheme and url_parts.netloc:
+                new_url_parts = list(url_parts)
+                new_url_parts[0] = url_parts.scheme.lower()
+                new_url_parts[1] = url_parts.netloc.lower()
+                return urlunparse(new_url_parts)
+
+            raise TypeError('URL: %s is not a valid URL format' % value)
+
 
 class EntityField(AdaptableTypeField):
     """
     :class:`Field` that can contain an :class:`Entity`
     """
+
+    # A map storing resolved dotted-name class types
+    _dotted_name_types = {}
+
     def __init__(self, entity, *args, **kwargs):
         """
         :param entity: The :class:`Entity` class to expect for this field.
                        Use 'self' to use the :class:`Entity` class that
                        this field is already bound to.
         """
-        self.type = entity
+        self._type = entity
         super(EntityField, self).__init__(*args, **kwargs)
 
+    @staticmethod
+    def _resolve_dotted_name(dotted_name):
+        try:
+            if '.' in dotted_name:
+                modules, _kls_name = dotted_name.rsplit('.', 1)
+                _module = __import__(modules, fromlist=[_kls_name])
+                _kls = getattr(_module, _kls_name)
+                assert callable(_kls), 'Dotted-name entity types should be callable.'
+                return _kls
+            elif 'self' != dotted_name:
+                raise ValueError('Invalid class name for EntityField: %s' % dotted_name)
+        except Exception:
+            raise_(
+                ValueError,
+                ValueError('Invalid class name for EntityField: %s' % dotted_name),
+                sys.exc_info()[2]
+            )
+
+    @property
+    def type(self):
+        """
+        Determine the type of the Entity that will be instantiated.
+
+        There are three ways to reference an Entity when using an EntityField:
+
+        - 'self': A byte string referencing the class that is defining this
+            EntityField as an attribute.
+        - '{dotted.name.kls}': A byte string referencing an importable callable
+            that can be instantiated at field-instantiation time.
+        - {Entity}: A type that subclasses `Entity`.
+
+        The order of operations during instantiation and resolution of the
+        above references is important; During the creation of an `Entity`, the
+        metaclass will call `init()` for fields defined on the class. This
+        is useful for the 'self' reference so that the `EntityField` can
+        be initialized with the class that is being instantiated during
+        creation of the instance. For dotted-name class strings, this is too
+        early since the dotted-name reference may not exist yet. For this
+        reason, resolving the dotted-name reference is deferred to be as late
+        as possible, in this case on the first read of the `type` property of
+        this field.
+
+        The dotted-name references are stored in a map on the `EntityField`
+        class to prevent resolving and importing the dotted-name on every
+        instance of this `EntityField`.
+
+        Returns:
+            `type`: A type to use when instantiating the Entity for this
+                EntityField.
+
+        """
+        if isinstance(self._type, (bytes, text_type)):
+            if self._type not in self.__class__._dotted_name_types:
+                _kls = self._resolve_dotted_name(self._type)
+                self.__class__._dotted_name_types[self._type] = _kls
+            return self.__class__._dotted_name_types[self._type]
+
+        return self._type
+
     def init(self, cls):
-        if self.type == 'self':
-            self.type = cls
+        if self._type == 'self':
+            self._type = cls
 
     def flatten(self, value):
         """
@@ -383,6 +562,7 @@ class EntityField(AdaptableTypeField):
         if value is not None:
             return value.jsonify()
 
+
 class IdField(Field):
     """
     A :class:`Field` that is used as the primary identifier for an :class:`Entity`
@@ -390,6 +570,7 @@ class IdField(Field):
     TODO This should accept another Field type to contain the ID
     """
 
+
 class CollectionField(Field):
     """
     A :class:`Field` that can contain an ordered list of values matching
@@ -448,13 +629,14 @@ class CollectionField(Field):
 _type_map = {
     datetime: DateTimeField(),
     int: IntField(),
-    basestring: StringField(),
-    unicode: StringField(),
+    bytes: BytesField(),
     str: StringField(),
+    text_type: StringField(),
     float: FloatField(),
     bool: BooleanField(),
 }
 
+
 def get_field_for_type(obj):
     t = type(obj)
     if t in _type_map:
